"""
Extension library for surface diffraction.

Rebuild of old C++ extension module.
"""

from logging import debug

import numba

from numpy import complex128, exp, pi, zeros


@numba.jit(
    numba.complex128[:](
        numba.float64[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64[:],
        numba.float64[:],
        numba.complex128[:, :],
        numba.float64[:, :, :],
        numba.float64[:],
    ),
    nopython=True,
    parallel=True,
    cache=True,
)
def surface_lattice_sum(x, y, z, h, k, l, u, oc, f, Pt, dinv):
    Nh = h.shape[0]
    Noc = oc.shape[0]
    fs = zeros(Nh, dtype=complex128)

    for i in numba.prange(Nh):
        pidi = -2.0 * (pi * dinv[i]) ** 2
        hi, ki, li = h[i], k[i], l[i]
        for j in range(Noc):
            # Loop over symmetry operations
            tmp = 0.0j
            for m in range(Pt.shape[0]):
                tmp += exp(
                    2.0j
                    * pi
                    * (
                        hi * (Pt[m, 0, 0] * x[j] + Pt[m, 0, 1] * y[j] + Pt[m, 0, 2])
                        + ki * (Pt[m, 1, 0] * x[j] + Pt[m, 1, 1] * y[j] + Pt[m, 1, 2])
                        + li * z[j]
                    )
                )
            fs[i] += oc[j] * f[i, j] * exp(pidi * u[j]) * tmp
    return fs



"""
==========================================================
:mod:`resolution` Data loader that includes the resoultion
==========================================================

Loads files in a four column format where the fourth column contains the resolution of the experiment.
Note that all reflectivity modules uses the standard deviation as a the resolution. Some instruments might
use the FWHM instead.

The default columns are the following:

* First column q values
* Second column Intensitiy values
* Third values The uncertainty in the Intensities
* Fourth column resolution

The other settings are just as in the default data loader.

The resolution is stored as the member variable res. For data set 0 it can accessed as
data[0].res
"""

import numpy as np

from ..data_loader_framework import Template
from ..utils import ShowWarningDialog

try:
    import wx
except ImportError:

    class void:
        pass

    wx = void()
    wx.Dialog = void


class Plugin(Template):
    def __init__(self, parent):
        Template.__init__(self, parent)
        self.q_col = 0
        self.I_col = 1
        self.eI_col = 2
        self.res_col = 3
        self.comment = "#"
        self.skip_rows = 0
        self.delimiter = None

    def LoadData(self, dataset, filename, data_id=0):
        """LoadData(self, dataset, filename) --> none

        Loads the data from filename into the data_item_number.
        """
        try:
            with open(filename, encoding="utf-8", errors="ignore") as fh:
                load_array = np.loadtxt(fh, delimiter=self.delimiter, comments=self.comment, skiprows=self.skip_rows)
        except Exception as e:
            ShowWarningDialog(
                self.parent,
                "Could not load the file: "
                + filename
                + " \nPlease check the format.\n\n numpy.loadtxt"
                + " gave the following error:\n"
                + str(e),
            )
        else:
            # For the freak case of only one data point
            if len(load_array.shape) < 2:
                load_array = np.array([load_array])
            # Check so we have enough columns
            if load_array.shape[1] - 1 < max(self.q_col, self.res_col, self.I_col, self.eI_col):
                ShowWarningDialog(
                    self.parent,
                    "The data file does not contain"
                    + "enough number of columns. It has "
                    + str(load_array.shape[1])
                    + " columns. Rember that the column index start at zero!",
                )
                # Okay now we have showed a dialog lets bail out ...
                return
            # The data is set by the default Template.__init__ function, neat hu
            # Note that the loaded data goes into *_raw so that they are not
            # changed by the transforms

            # print load_array
            dataset.x_raw = load_array[:, self.q_col]
            dataset.y_raw = load_array[:, self.I_col]
            dataset.error_raw = load_array[:, self.eI_col]
            dataset.set_extra_data("res", load_array[:, self.res_col], "res")
            # Run the commands on the data - this also sets the x,y, error memebers
            # of that data item.
            dataset.run_command()

            # insert metadata into ORSO compatible fields
            dataset.meta["data_source"]["facility"] = "GenX resolution data loader"
            dataset.meta["data_source"]["experiment"]["instrument"] = "unkown"
            dataset.meta["data_source"]["import_columns"] = dict(
                x=self.q_col, y=self.I_col, error=self.eI_col, res=self.res_col
            )

    def SettingsDialog(self):
        """SettingsDialog(self) --> None

        This function should - if necessary implement a dialog box
        that allows the user set import settings for example.
        """
        col_values = {"I": self.I_col, "q": self.q_col, "Resolution": self.res_col, "I error": self.eI_col}
        misc_values = {"Comment": str(self.comment), "Skip rows": self.skip_rows, "Delimiter": str(self.delimiter)}
        dlg = SettingsDialog(self.parent, col_values, misc_values)
        if dlg.ShowModal() == wx.ID_OK:
            col_values = dlg.GetColumnValues()
            misc_values = dlg.GetMiscValues()
            self.q_col = col_values["q"]
            self.res_col = col_values["Resolution"]
            self.I_col = col_values["I"]
            self.eI_col = col_values["I error"]
            self.comment = misc_values["Comment"]
            self.skip_rows = misc_values["Skip rows"]
            self.delimiter = misc_values["Delimiter"]
            self.SetStatusText("Changed import settings")
        else:
            self.SetStatusText("No change to import settings")
        dlg.Destroy()


class SettingsDialog(wx.Dialog):

    def __init__(self, parent, col_values, misc_values):
        wx.Dialog.__init__(self, parent, -1, "Data loader settings")

        box_sizer = wx.BoxSizer(wx.HORIZONTAL)

        # Make the box for putting in the columns
        col_box = wx.StaticBox(self, -1, "Columns")
        col_box_sizer = wx.StaticBoxSizer(col_box, wx.VERTICAL)

        # col_values = {'y': 1,'x': 0,'y error': 1}
        col_grid = wx.GridBagSizer(len(col_values), 2)
        self.col_controls = col_values.copy()
        keys = list(col_values.keys())
        keys.sort()
        for i, name in enumerate(keys):
            text = wx.StaticText(self, -1, name + ": ")
            control = wx.SpinCtrl(self)
            control.SetRange(-1, 100)
            control.SetValue(col_values[name])
            col_grid.Add(text, (i, 0), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
            col_grid.Add(control, (i, 1), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
            self.col_controls[name] = control

        col_box_sizer.Add(col_grid, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        box_sizer.Add(col_box_sizer, 0, wx.ALL | wx.EXPAND, 5)

        col_box = wx.StaticBox(self, -1, "Misc")
        col_box_sizer = wx.StaticBoxSizer(col_box, wx.VERTICAL)

        # Lets add another box for comments and rows to skip
        # misc_values = {'Comment': '#', 'Skip rows': 0,'Delimiter': 'None'}
        col_grid = wx.GridBagSizer(len(misc_values), 2)
        self.misc_controls = misc_values.copy()
        keys = list(misc_values.keys())
        keys.sort()
        for i, name in enumerate(keys):
            text = wx.StaticText(self, -1, name + ": ")
            if type(misc_values[name]) == type(1):
                control = wx.SpinCtrl(self)
                control.SetRange(0, 100)
                control.SetValue(misc_values[name])
            else:
                control = wx.TextCtrl(self, value=misc_values[name], style=wx.EXPAND)
            col_grid.Add(text, (i, 0), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
            col_grid.Add(control, (i, 1), flag=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL, border=5)
            self.misc_controls[name] = control

        col_box_sizer.Add(col_grid, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
        box_sizer.Add(col_box_sizer, 0, wx.ALL | wx.EXPAND, 5)

        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        button_sizer.Realize()

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(box_sizer, 1, wx.GROW, 20)
        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW, 30)

        sizer.Add(button_sizer, 0, flag=wx.ALIGN_RIGHT, border=20)
        self.SetSizer(sizer)

        sizer.Fit(self)
        self.Layout()

    def GetColumnValues(self):
        values = {}
        for key in self.col_controls:
            values[key] = self.col_controls[key].GetValue()
        return values

    def GetMiscValues(self):
        values = {}
        for key in self.misc_controls:
            val = self.misc_controls[key].GetValue()
            if type(val) == type("") or type(val) == type(""):
                if val.lower() == "none":
                    val = None
            values[key] = val
        return values



r"""
.. module:: materials
   :synopsis: Classes to work with differnt materials in scattering experiments

"""

import pickle
import re

from os import path

import numpy as np

from scipy import interpolate

from . import scatteringlengths as _sl
from .parameters import Calc, ComplexArray, Float, HasParameters
from .physical_constants import r_e, AA_to_eV

_module_dir, _filename = path.split(__file__)
# Look only after the file name and not the ending since
# the file ending can be pyc if compiled...
_filename = _filename.split(".")[0]
_module_dir = "." if _module_dir == "" else _module_dir

_atomic_weights = _sl.load_atomic_weights_dabax(path.join(_module_dir, "../databases/AtomicWeights.dat"))


def create_fpdisp_lookup(path):
    """Creates a lookup function to lookup element names and returns a function
    that yields a function of dispersive scattering factors f(E) at Q = 0. NOTE energy is in eV
    """

    def create_dispersion_func(name):
        e, f1, f2 = np.loadtxt(path + "%s.nff" % name.lower(), skiprows=1, unpack=True)
        f1interp = interpolate.interp1d(e, f1, kind="linear")
        f2interp = interpolate.interp1d(e, f2, kind="linear")
        def_wl = _AA_to_eV / np.mean(e)

        @MemoizeF
        def f(wl=def_wl, **kwargs):
            return f1interp(_AA_to_eV / wl) - 1.0j * f2interp(_AA_to_eV / wl)

        return Calc(f)

    return create_dispersion_func


f0 = create_fpdisp_lookup(path.join(_module_dir, "../databases/f1f2_nist/"))


class MemoizeF:
    """Remember previous evaluations of function fn and stores it in a dictionary.

    Based on the Cookbook recipe: Memoizing (Caching) the Return Values of Functions in
    Python Cookbook by David Ascher, Alex Martelli.
    """

    def __init__(self, fn):
        self.fn = fn
        self.memo = {}

    def __call__(self, wl=1.54, **kwds):
        pickled = pickle.dumps(wl, 1)
        if pickled not in self.memo:
            self.memo[pickled] = self.fn(wl=wl, **kwds)
        return self.memo[pickled]


class Material(HasParameters):
    """Class to keep transform a chemical formula and mass density to scattering length

    The formula is given in ASCII format, for example 'Ag2(NO2)3' or 'HNO3', nested parentheses are allowed.
    Also non-integer indices are allowed for example Fe0.01Pd

    """

    validation_kwargs = {"wl": 1.54}

    # The scattering length should be in AA-2
    # Need to transform g/cm3 to formula units per AA3
    # 1u = 1.66054×10−27 kg 1 AA=10−10m
    # density[kg/m3] = 1.66054*10^3*density[u/A3]
    # density [g/cm3] = density [kg/m3] * 1e-3/1e-2**3 = density [kg/m3] * 1e3
    # So formula units per AA^ = density [kg/m3]/1.66054*10^3/weight per formula unit
    # or formula units per AA^ = density [g/cm3]/1.66054/weight per formula unit

    def __init__(self, formula, **kwargs):
        self.sld_x = ComplexArray(0, unit="1/AA-2", help="Scattering length density")
        self.density = Float(0.0, unit="g/cm3", help="Weight density")
        self._formula_density = Float(0.0, unit="u", help="weight of one formula unit")
        HasParameters.__init__(self, **kwargs)
        self.formula_dict = {}
        self.parse_formula(formula)

        # Create the composition variables needed to describe formula
        w_list = []
        f0_list = []
        for name in self.formula_dict:
            self.__setattr__(name, Float(self.formula_dict[name], help="Number of %s atoms" % name))
            element_count = self.__getattribute__(name)
            w_list.append(element_count * self.get_atomic_weight(name))
            f0_list.append(element_count * self.get_f0(name))
        self._formula_density = self.density / 1.66054 / sum(w_list)
        self.sld_x = self._formula_density * sum(f0_list) * r_e

    def get_atomic_weight(self, element):
        """Returns the atomic weight in unified atomic mass units (u)

        Parameters:
            element(string): name of the element

        """
        return _atomic_weights[element.lower()]

    def get_f0(self, element):
        """Returns the complex atomic forward scattering length normalised to the scattering length of one electron for
         element.

        Parameters:
            element(string): name of the element

        Returns:
            f0(function): function for f0 that takes the wavelength as a keyword argument
        """
        return f0(element)

    def parse_formula(self, formula, multiplier=1):
        """Parses the string formula and returns a list of isotopes

        Parameters:
            formula(string): A string containing the formula.
            multiplier(int): A number representing that will be used to multiply the element count
        """
        # regexp = re.compile(r"([A-Z][a-z]?)(\d*)|\((.*)\)(\d*)")
        regexp = re.compile(r"([A-Z][a-z]?)(\d*\.?\d*)|\((.*)\)(\d*\.?\d*)")

        element, element_count, group, group_count = (1, 2, 3, 4)
        pos = 0
        while pos < len(formula):
            match = regexp.match(formula, pos)
            if not match:
                raise ValueError("Error in formula: %s at position %d" % (formula, pos))
            pos = match.end()
            if match.group(element):
                # print "Match %s is an element" % match.group(0)
                count = self.formula_count(match.group(element_count), multiplier)
                try:
                    self.formula_dict[match.group(element)] += count
                except KeyError:
                    self.formula_dict[match.group(element)] = count
            elif match.group(group):
                # print "Group found: %s, count: %s" % (match.group(group), match.group(group_count))
                count = self.formula_count(match.group(group_count), multiplier)
                self.parse_formula(match.group(group), multiplier=count)

    def formula_count(self, count_str, multiplier):
        """Handles the parsing result for the count of an formula part. Returns the count*multiplier"

        Parameters:
            count_str(string): The count of the element/group as a string
            multiplier(float): The  multiplier from previous groups.

        """
        if not count_str:
            # Set count to 1 if count is not exist, ie a single group is parsed
            count = 1
        else:
            count = float(count_str)
        return count * multiplier



"""
Test of special list class in colors module.
"""

import os
import unittest

from dataclasses import dataclass, field, fields

from genx.core import config as c
from genx.exceptions import GenxIOError, GenxOptionError

CFG_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "test.conf")
EXP_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "export.conf")


class TestConfig(unittest.TestCase):

    def test_load_default(self):
        config = c.Config()
        config.load_default(CFG_PATH, reset=True)

        # check that the default_config data was actually loaded and the model_config reset
        self.assertEqual(config.default_config["section_1"], {"p int": "1", "p float": "1.1", "p str": "hallo world"})
        self.assertEqual(config.model_config.sections(), [])

        with self.assertRaises(GenxIOError):
            config.load_default("none existant path")

    def test_write_default(self):
        config = c.Config()
        config.write_default(EXP_PATH)
        self.assertTrue(os.path.isfile(EXP_PATH))
        os.unlink(EXP_PATH)

    def test_fget(self):
        config = c.Config()
        config.default_config["section_1"] = {
            "p int": "1",
            "p float": "1.1",
            "p str": "hallo world",
            "p bool": True,
            "p2 bool": 0,
        }
        config.model_config["section_1"] = {"p int": "2", "p float": "2.2", "p str": "hallo local"}

        # items from local config
        self.assertEqual(config._getf("getint", "section_1", "p int"), 2)
        self.assertTrue(isinstance(config._getf("getint", "section_1", "p int"), int))
        self.assertEqual(config._getf("getfloat", "section_1", "p float"), 2.2)
        self.assertTrue(isinstance(config._getf("getfloat", "section_1", "p float"), float))
        self.assertEqual(config._getf("get", "section_1", "p str"), "hallo local")
        # items from default config
        self.assertEqual(config._getf("getboolean", "section_1", "p bool"), True)
        self.assertEqual(config._getf("getboolean", "section_1", "p2 bool"), False)
        self.assertTrue(isinstance(config._getf("getboolean", "section_1", "p bool"), bool))

        # conversion failing
        self.assertEqual(config._getf("getboolean", "section_1", "p str", fallback=True), True)

        with self.assertRaises(AttributeError):
            config._getf("getbeer", "section_1", "p bool")
        with self.assertRaises(GenxOptionError):
            config._getf("getint", "section_2", "p bool")
        with self.assertRaises(GenxOptionError):
            config._getf("getint", "section_1", "p bool")
        with self.assertRaises(GenxOptionError):
            config._getf("getint", "section_1", "not existing")

    def test_get_other(self):
        config = c.Config()
        config.default_config["section_1"] = {
            "p int": "1",
            "p float": "1.1",
            "p str": "hallo world",
            "p bool": True,
            "p2 bool": 0,
        }
        config.model_config["section_1"] = {"p int": "2", "p float": "2.2", "p str": "hallo local", "p list": "1;2;3"}

        self.assertEqual(config.getlist("section_1", "p list"), ["1", "2", "3"])
        with self.assertRaises(AttributeError):
            config.not_existant

    def test_read_string(self):
        config = c.Config()
        config.load_string("[section_1]\nbier = 13")
        self.assertEqual(config.model_config["section_1"]["bier"], "13")

        with self.assertRaises(GenxIOError):
            config.load_string("not the right format")

    def test_set(self):
        config = c.Config()
        config.set("section_1", "p int", 1)
        config.default_set("section_1", "p int", 2)
        self.assertEqual(config.model_config["section_1"]["p int"], "1")
        self.assertEqual(config.default_config["section_1"]["p int"], "2")

    def test_dump(self):
        config = c.Config()
        config.set("section_1", "p int", 1)
        config.set("section_1", "p float", 1.5)
        dump = config.model_dump()
        config.set("section_1", "p int", 2)
        self.assertEqual(config.getint("section_1", "p int"), 2)
        config.load_string(dump)
        self.assertEqual(config.getint("section_1", "p int"), 1)
        config.load_string(dump)


class TestBaseConfig(unittest.TestCase):

    def setUp(self):
        @dataclass
        class TestConfig(c.BaseConfig):
            section = "section_1"
            p_int: int
            p_float: float = 1.1
            p_str: str = "hallo world"
            p_bool: bool = True
            p_list: list = field(default_factory=list)

        self.test_class = TestConfig

    def test_abc(self):
        class Broken(c.BaseConfig): ...

        with self.assertRaises(TypeError):
            Broken()

    def test_load_default(self):
        cfg = self.test_class(p_int=1)
        self.assertEqual(
            cfg.asdict(), {"p_bool": True, "p_float": 1.1, "p_int": 1, "p_list": [], "p_str": "hallo world"}
        )

        c.config = c.Config()
        c.config.model_set("section_1", "p int", "13")
        c.config.model_set("section_1", "p float", "13.13")
        c.config.model_set("section_1", "p bool", "false")
        c.config.model_set("section_1", "p str", "hallo local")
        c.config.model_set("section_1", "p list", "1;2;3")
        cfg.load_config()
        self.assertEqual(
            cfg.asdict(),
            {"p_bool": False, "p_float": 13.13, "p_int": 13, "p_list": ["1", "2", "3"], "p_str": "hallo local"},
        )
        self.assertEqual(cfg.p_int, 13)
        self.assertEqual(cfg.p_bool, False)
        self.assertEqual(cfg.p_float, 13.13)
        self.assertEqual(cfg.p_list, ["1", "2", "3"])
        self.assertEqual(cfg.p_str, "hallo local")
        # ignore wrong values with a debug message
        c.config.model_set("section_1", "p int", "bier")
        cfg.load_config()
        self.assertEqual(cfg.p_int, 13)

    def test_save(self):
        cfg = self.test_class(p_int=1)
        c.config = c.Config()
        cfg.save_config(default=False)
        self.assertEqual(
            dict(c.config.model_config["section_1"]),
            {"p int": "1", "p float": "1.1", "p str": "hallo world", "p bool": "True", "p list": ""},
        )
        self.assertEqual(c.config.default_config.sections(), [])
        cfg.save_config(default=True)
        self.assertEqual(
            dict(c.config.default_config["section_1"]),
            {"p int": "1", "p float": "1.1", "p str": "hallo world", "p bool": "True", "p list": ""},
        )

    def test_copy(self):
        cfg = self.test_class(p_int=1)
        cpy = cfg.copy()
        self.assertEqual(cpy, cfg)
        self.assertFalse(cpy is cfg)

    def test_groups(self):
        cfg = self.test_class(p_int=1)
        # check that the items match, order could be arbitrary
        self.assertEqual(set(cfg.groups[""]), set(["p_bool", "p_float", "p_int", "p_list", "p_str"]))

    def test_fields(self):
        cfg = self.test_class(p_int=1)
        flds = list(fields(cfg))
        self.assertEqual(cfg.get_fields(), flds)
        self.assertEqual(len(cfg.get_fields()), 5)
        self.assertEqual(len(cfg.get_fields(["p_bool", "p_int"])), 2)
        self.assertEqual(len(cfg.get_fields(["p_bool", "none"])), 1)

    def test_gparm(self):
        res = self.test_class.GParam(13, pmin=53, pmax=12, label="hello world", descriptoin="this is a parameter")
        self.assertEqual(res.default, 13)
        self.assertEqual(
            res.metadata["genx"], {"pmin": 53, "pmax": 12, "label": "hello world", "descriptoin": "this is a parameter"}
        )

        res = self.test_class.GChoice(13, [13, 14, 15], label="hello world", descriptoin="this is a parameter")
        self.assertEqual(res.default, 13)
        self.assertEqual(
            res.metadata["genx"],
            {"selection": [13, 14, 15], "label": "hello world", "descriptoin": "this is a parameter"},
        )

    def test_merged(self):
        @dataclass
        class TestSecond(c.BaseConfig):
            section = "section_2"
            second: int = 14
            groups = {"second": ["second"]}

        cfg = self.test_class(p_int=1)
        cfg2 = TestSecond()
        mrg = cfg | cfg2
        mrg.p_float = 13.13
        self.assertEqual(cfg.p_float, 13.13)
        self.assertEqual(mrg.p_float, 13.13)
        with self.assertRaises(AttributeError):
            mrg.does_not_exist

        self.assertEqual(set(mrg.groups[""]), set(["p_bool", "p_float", "p_int", "p_list", "p_str"]))
        self.assertEqual(set(mrg.groups["second"]), set(["second"]))

        self.assertEqual(mrg.get_fields(), list(fields(cfg)) + list(fields(cfg2)))

        cpy = mrg.copy()
        self.assertNotEqual(cfg, mrg)
        self.assertEqual(cpy, mrg)
        self.assertFalse(cpy is mrg)

        self.assertEqual(len((mrg | mrg)._children), 4)
        self.assertEqual(len((mrg | cfg)._children), 3)


class TestConfigurable(unittest.TestCase):

    def test_subclassing(self):
        @dataclass
        class Config(c.BaseConfig):
            section = "section_1"
            test: str = "start value"

        class MyConfigurable(c.Configurable):
            opt: Config

        cls = MyConfigurable()
        cls.WriteConfig()
        cls.opt.test = "changed value"
        self.assertEqual(cls.opt.test, "changed value")
        cls.ReadConfig()
        self.assertEqual(cls.opt.test, "start value")

    def test_missing_opt(self):
        class MyConfigurable(c.Configurable): ...

        with self.assertRaises(TypeError):
            MyConfigurable()
        with self.assertRaises(ValueError):
            MyConfigurable(config_class=str)

        @dataclass
        class Config(c.BaseConfig):
            section = "section_1"
            test: str = "start value"

        MyConfigurable(config_class=Config)



"""
Helper module to support ORSO file header information to build reflectivity models.
"""
import logging

from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional

from orsopy import fileio

from ...add_ons.help_modules.materials_db import MASS_DENSITY_CONVERSION, Formula

if TYPE_CHECKING:
    # make sure the wx based plugins don't need to be imported at runtime
    from ...add_ons.Reflectivity import Plugin as RPlugin
    from ...add_ons.SimpleReflectivity import Plugin as SRPlugin


def sanetize_name(name):
    if name:
        name = name.replace('\t', '_').replace(' ', '_').replace('.', 'p').replace('-', 'm')
        if not name.isidentifier():
            name = None
    return name


@dataclass
class InstrumentInformation:
    probe: str
    coords: str
    wavelength: float


@dataclass
class LayerData:
    dens: float
    formula: Optional[Formula]
    f: complex
    b: complex
    d: float
    sigma: float
    name: Optional[str]

    def __post_init__(self):
        self.name = sanetize_name(self.name)

@dataclass
class StackData:
    layers: List[LayerData]
    name: Optional[str]

    def __post_init__(self):
        self.name = sanetize_name(self.name)

@dataclass
class LayerModel:
    substrate: LayerData
    ambient: LayerData
    stacks: List[StackData]
    repetitions: List[int]

class OrsoHeaderAnalyzer:
    header: fileio.Orso
    instrument: InstrumentInformation
    layer_model: Optional[LayerModel] = None

    def __init__(self, meta):
        self.model = None
        meta = deepcopy(meta)
        self.header = fileio.Orso.from_dict(meta)
        self.analyze_instrument_info()
        if hasattr(fileio, "model_language") and self.header.data_source.sample.model:
            try:
                self.analyze_model()
            except Exception:
                logging.error('Header defined a sample model but parsing the model failed', exc_info=True)

    @classmethod
    def from_orso(cls, orso_header):
        self = object.__new__(cls)
        self.model = None
        self.header = orso_header
        self.analyze_instrument_info()
        try:
            self.analyze_model()
        except Exception:
            logging.critical('Header defined a sample model but parsing the model failed', exc_info=True)
        return self

    @classmethod
    def from_yaml(cls, yaml_text):
        import yaml
        from orsopy.fileio.model_language import SampleModel

        meta = yaml.safe_load(yaml_text)

        for parent in ['data_source', 'sample', 'model']:
            if parent in meta:
                meta = meta[parent]
        if not 'stack' in meta:
            raise ValueError('A sample model must contain the "stack" attribute')

        self = object.__new__(cls)
        self.model = None
        self.header = fileio.Orso.empty()
        self.header.data_source.sample.model = SampleModel.from_dict(meta)

        self.instrument = InstrumentInformation(probe='x-ray', coords='q', wavelength=1.54)

        try:
            self.analyze_model()
        except Exception:
            logging.critical('Header defined a sample model but parsing the model failed', exc_info=True)

        return self

    @property
    def instrument_settings(self) -> fileio.InstrumentSettings:
        return self.header.data_source.measurement.instrument_settings

    @property
    def experiment(self) -> fileio.Experiment:
        return self.header.data_source.experiment

    def get_probe(self):
        if self.experiment.probe == "neutron":
            pol = self.instrument_settings.polarization or "unpolarized"
            if pol == "unpolarized":
                probe = "neutron"
            elif pol in ["pp", "mm", "pm", "mp"]:
                probe = "neutron pol spin flip"
            else:
                probe = "neutron pol"
        else:
            probe = "x-ray"
        return probe

    def analyze_instrument_info(self):
        probe = self.get_probe()

        # detect x-axis unit
        if self.header.columns[0].unit in [None, "1/angstrom", "1/nm"]:
            coords = "q"
        else:
            # not officially ORSO confirm
            coords = "2θ"
        try:
            wavelength = float(self.instrument_settings.wavelength.as_unit("angstrom"))
        except Exception:
            wavelength = 1.54
        if not isinstance(wavelength, float):
            # in case wavelength is a ValueRange (ToF)
            wavelength = 1.0
        self.instrument = InstrumentInformation(probe=probe, coords=coords, wavelength=wavelength)

    @staticmethod
    def get_layer_data(layer) -> LayerData:
        layer.material.generate_density()
        if getattr(layer.material, 'formula', None):
            formula = Formula.from_str(layer.material.formula)
            if layer.material.number_density:
                dens = layer.material.number_density.as_unit("1/angstrom**3")
            elif layer.material.mass_density:
                dens = layer.material.mass_density.as_unit("g/cm**3") * MASS_DENSITY_CONVERSION / formula.mFU()
            else:
                from genx.models.utils import bc

                try:
                    dens = layer.material.get_sld().real*1e5 / eval(formula.b(), globals={'bc': bc}, locals={}).real
                except Exception:
                    dens = 0.1
            b = formula.b()
            f = formula.f()
        else:
            formula = None
            b = layer.material.get_sld()*1e6
            f = layer.material.get_sld()*1e6
            dens = 0.1
        name = getattr(layer, 'original_name', None) or getattr(layer.material, 'original_name', None)
        d = layer.thickness.as_unit("angstrom")
        sigma = layer.roughness.as_unit("angstrom")
        return LayerData(dens=dens, formula=formula, b=b, f=f, d=d, sigma=sigma, name=name)

    def analyze_model(self):
        from orsopy.fileio import model_language

        header_model = self.header.data_source.sample.model

        res_stack = header_model.resolve_stack()
        stacks = []
        repetitions = []
        last_stack = True
        for si in res_stack:
            if hasattr(si, "repetitions"):
                repetitions.append(si.repetitions)
                stack = StackData(layers=[], name=getattr(si, 'original_name', None))
                for stack_item in si.sequence:
                    if isinstance(stack_item, model_language.Layer):
                        stack.layers.append(self.get_layer_data(stack_item))
                    else:
                        stack.layers += list(map(self.get_layer_data, stack_item.resolve_to_layers()))
                stacks.append(stack)
                last_stack = True
            elif last_stack:
                last_stack = False
                repetitions.append(1)
                stacks.append(StackData(layers=[self.get_layer_data(si)], name=None))
            else:
                stacks[-1].layers.append(self.get_layer_data(si))

        # make sure we remove ambient and substrate from a single repetition stack
        if repetitions[0] != 1:
            repetitions.insert(0, 1)
            repetitions[1] -= 1
            stacks.insert(0, stacks[0])
        if repetitions[-1] != 1:
            repetitions[-1] -= 1
            repetitions.append(1)
            stacks.append(stacks[-1])
        ambient = stacks[0].layers.pop(0)
        substrate = stacks[-1].layers.pop(-1)
        for i, si in reversed(list(enumerate(stacks))):
            if len(si.layers) == 0:
                stacks.pop(i)
                repetitions.pop(i)
        self.layer_model = LayerModel(ambient=ambient, substrate=substrate, stacks=stacks, repetitions=repetitions)

    def simple_refl_layer(self, layer: LayerData, pos=1):
        if layer.formula:
            dens = layer.dens * layer.formula.mFU() / MASS_DENSITY_CONVERSION
            return [
                layer.name,
                "Formula",
                layer.formula,
                False,
                str(dens),
                False,
                "0.0",
                False,
                str(layer.d),
                False,
                str(layer.sigma),
                pos,
            ]
        else:
            return [
                layer.name,
                "Formula",
                "SLD",
                False,
                str(layer.b.real if self.instrument.probe != "x-ray" else layer.f.real),
                False,
                "0.0",
                False,
                str(layer.d),
                False,
                str(layer.sigma),
                pos,
            ]

    def build_simple_model(self, refl: "SRPlugin"):
        refl.sample_widget.sample_table.ResetModel()
        refl.sample_widget.inst_params["probe"] = self.instrument.probe
        refl.sample_widget.inst_params["wavelength"] = self.instrument.wavelength
        refl.sample_widget.inst_params["coords"] = self.instrument.coords
        refl.sample_widget.inst_params["res"] = 0.01
        self.build_simple_sample(refl)

    def build_simple_sample(self, refl: "SRPlugin"):
        from ...add_ons.SimpleReflectivity import BOT_LAYER, ML_LAYER, TOP_LAYER
        if self.layer_model:
            # if any stack is repeated, use that as central one
            if any([ri>1 for ri in self.layer_model.repetitions]):
                pos = TOP_LAYER
                is_ML = True
            else:
                pos = ML_LAYER
                is_ML = False

            ID = 0
            layers = []
            repetitions = 1
            for i, (ri, si) in enumerate(zip(self.layer_model.repetitions, self.layer_model.stacks)):
                if is_ML and pos==ML_LAYER:
                    pos = BOT_LAYER
                if ri>1 and pos==TOP_LAYER:
                    pos = ML_LAYER
                    repetitions = ri
                    for li in si.layers:
                        nl = self.simple_refl_layer(li, pos=pos)
                        if nl[0] is None:
                            nl[0] = f"Layer_{ID:02}"
                        elif nl[0] in [ni[0] for ni in layers]:
                            nl[0] = f"{nl[0]}_{ID:02}"
                        layers.append(nl)
                        ID += 1
                else:
                    for li in ri*si.layers:
                        nl = self.simple_refl_layer(li, pos=pos)
                        if nl[0] is None:
                            nl[0] = f"Layer_{ID:02}"
                        elif nl[0] in [ni[0] for ni in layers]:
                            nl[0] = f"{nl[0]}_{ID:02}"
                        layers.append(nl)
                        ID += 1
            refl.sample_widget.sample_table.ambient = self.simple_refl_layer(self.layer_model.ambient)
            refl.sample_widget.sample_table.substrate = self.simple_refl_layer(self.layer_model.substrate)
            refl.sample_widget.sample_table.RebuildTable(layers)
            refl.sample_widget.sample_table.repetitions = repetitions
        refl.sample_widget.last_sample_script = refl.sample_widget.sample_table.getModelCode()
        refl.sample_widget.UpdateModel(re_color=True)

    def build_reflectivity(self, refl: "RPlugin"):
        refl.CreateNewModel("models.spec_nx")
        # detect source radiation
        inst = refl.sample_widget.instruments["inst"]
        inst.probe = self.instrument.probe
        inst.coords = self.instrument.coords
        inst.wavelength = self.instrument.wavelength
        pol_names = {"po": "uu", "mo": "dd", "op": "uu", "om": "dd", "pp": "uu", "mm": "dd", "pm": "ud", "mp": "du"}
        # set resolution column
        if len(self.header.columns) > 3 and self.header.columns[3].error_of == self.header.columns[0].name:
            inst.restype = "full conv and varying res."
            inst.respoints = 7
            inst.resintrange = 2.5
            el = refl.simulation_widget.GetExpressionList()
            for i, data_item in enumerate(refl.GetModel().data):
                if len(data_item.meta["columns"]) > 3 and (
                    data_item.meta["columns"][3].get("error_of", None) == data_item.meta["columns"][0]["name"]
                ):
                    el[i].append(f"inst.setRes(data[{i}].res)")
                else:
                    el[i].append(f"inst.setRes(0.001)")
                # set polarization channel
                pol = data_item.meta["data_source"]["measurement"]["instrument_settings"].get(
                    "polarization", "unpolarized"
                )
                if pol in pol_names:
                    el[i].append(f'inst.setPol("{pol_names[pol]}")')

        self.build_sample(refl)

    def build_sample(self, refl: "RPlugin"):
        if self.layer_model:
            lm = self.layer_model
            tmp = refl.sampleh.sample.Ambient
            tmp.b, tmp.f, tmp.dens = lm.ambient.b, lm.ambient.f, lm.ambient.dens
            tmp._ca["f"] = lm.ambient.f
            tmp._ca["b"] = lm.ambient.b
            tmp = refl.sampleh.sample.Substrate
            tmp.b, tmp.f, tmp.dens, tmp.sigma = lm.substrate.b, lm.substrate.f, lm.substrate.dens, lm.substrate.sigma
            tmp._ca["f"] = lm.substrate.f
            tmp._ca["b"] = lm.substrate.b

            while len(refl.sampleh)>2:
                # clear the stacks in the sample handler
                refl.sampleh.deleteItem(1)

            pos = 1
            for si, (rep, stack) in enumerate(zip(lm.repetitions, lm.stacks)):
                name = stack.name
                if name is None:
                    name = f"Stack_{si:02}"
                elif name in refl.sampleh.names:
                    name = f"{name}_{si:02}"
                refl.sampleh.insertItem(pos, "Stack", name)
                stack_obj = refl.sampleh.sample.Stacks[0]
                stack_obj.Repetitions = rep
                pos += 1
                for li, layer in enumerate(stack.layers):
                    name = layer.name
                    if name is None:
                        name = f"Layer_{pos:02}"
                    elif name in refl.sampleh.names:
                        name = f"{name}_{pos:02}"

                    refl.sampleh.insertItem(pos, "Layer", name)
                    pos += 1

                    layer_obj = stack_obj.Layers[0]

                    layer_obj.b, layer_obj.f, layer_obj.dens = layer.b, layer.f, layer.dens
                    layer_obj.d, layer_obj.sigma = layer.d, layer.sigma
                    layer_obj._ca["b"] = layer.b
                    layer_obj._ca["f"] = layer.f

            refl.sample_widget.Update()

        refl.WriteModel()


# include GenX materials in density resolution
try:
    from orsopy.fileio import model_language
except ImportError:
    pass
else:
    from orsopy.utils.chemical_formula import Formula as OrsoFormula
    try:
        from orsopy.utils.density_resolver import DensityResolver
    except ImportError:
        from orsopy.utils.density_resolver import MaterialResolver as DensityResolver
    from orsopy.utils.resolver_slddb import ResolverSLDDB

    from ...add_ons.help_modules.materials_db import mdb

    class ResolverGenX(DensityResolver):
        comment = "from GenX "

        def resolve_formula(self, formula: OrsoFormula) -> float:
            frm = Formula.from_str(str(formula))
            if frm not in mdb:
                raise ValueError(f"Could not find material {formula}")
            else:
                res_formula, density = mdb[frm]
                self.comment = f"density from GenX db {res_formula}"
                return 1e3 * eval(density)

        def resolve_elemental(self, formula: OrsoFormula) -> float:
            raise ValueError("GenX resolve only used for known materials.")

    model_language.DENSITY_RESOLVERS = [ResolverGenX(), ResolverSLDDB()]



"""
An replacement for wx.SingleChoiceDialog that does not have issues with oversized width
and allows additional styling of elements.
"""

import wx


class SCDialog(wx.Dialog):
    def __init__(
        self,
        parent,
        message,
        caption,
        choices,
        background_colors=None,
        style=wx.CHOICEDLG_STYLE,
        pos=wx.DefaultPosition,
    ):
        wx.Dialog.__init__(self, parent, title=caption, style=style, pos=pos)
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        vbox.Add(wx.StaticText(self, label=message), 0, wx.FIXED_MINSIZE | wx.ALL, 4)
        self.list_widget = wx.ListBox(self, style=wx.LB_SINGLE | wx.LB_HSCROLL | wx.LB_OWNERDRAW, choices=choices)
        if background_colors:
            for i, ci in enumerate(background_colors):
                self.list_widget.SetItemBackgroundColour(i, ci)

        vbox.Add(self.list_widget, 1, wx.EXPAND)

        self.list_widget.Bind(wx.EVT_LISTBOX_DCLICK, self.selected)

    def selected(self, event):
        self.EndModal(wx.ID_OK)

    def GetSelection(self):
        return self.list_widget.GetSelection()



"""
GenX main package.
"""

# make sure the models can import subpackages correctly
import os
import sys

from importlib import import_module
from importlib.abc import MetaPathFinder
from importlib.util import spec_from_file_location

_buildin_metafinder = list(sys.meta_path)


class GenxModuleFinder(MetaPathFinder):

    def find_spec(self, fullname, path=None, target=None):
        if fullname.startswith("genx"):
            return None
        itms = fullname.split(".")
        fname = os.path.join(os.path.dirname(os.path.abspath(__file__)), *itms) + ".py"
        pname = os.path.join(os.path.dirname(os.path.abspath(__file__)), *itms, "__init__.py")
        if os.path.exists(fname):
            m = import_module("genx." + fullname)
            sys.modules[fullname] = m
            return spec_from_file_location(fullname, fname)
        if os.path.exists(pname):
            m = import_module("genx." + fullname)
            sys.modules[fullname] = m
            return spec_from_file_location(fullname, pname)


sys.meta_path.insert(0, GenxModuleFinder())



"""
Custom IDs for GenX menus and toolbars.
"""

from enum import Enum

import wx


class ToolId(int, Enum):
    NEW_MODEL = wx.Window.NewControlId()
    NEW_FROM_FILE = wx.Window.NewControlId()
    OPEN_MODEL = wx.Window.NewControlId()
    SAVE_MODEL = wx.Window.NewControlId()
    SIM_MODEL = wx.Window.NewControlId()
    START_FIT = wx.Window.NewControlId()
    STOP_FIT = wx.Window.NewControlId()
    RESTART_FIT = wx.Window.NewControlId()
    CALC_ERROR = wx.Window.NewControlId()
    SOLVER_SELECT = wx.Window.NewControlId()
    ZOOM = wx.Window.NewControlId()
    ERROR_STATS = wx.Window.NewControlId()


class MenuId(int, Enum):
    NEW_MODEL = wx.Window.NewControlId()
    NEW_FROM_FILE = wx.Window.NewControlId()
    OPEN_MODEL = wx.Window.NewControlId()
    SAVE_MODEL = wx.Window.NewControlId()
    SAVE_MODEL_AS = wx.Window.NewControlId()
    MODEL_BATCH = wx.Window.NewControlId()

    IMPORT_DATA = wx.Window.NewControlId()
    IMPORT_TABLE = wx.Window.NewControlId()
    IMPORT_SCRIPT = wx.Window.NewControlId()

    EXPORT_ORSO = wx.Window.NewControlId()
    EXPORT_DATA = wx.Window.NewControlId()
    EXPORT_TABLE = wx.Window.NewControlId()
    EXPORT_SCRIPT = wx.Window.NewControlId()

    PUBLISH_PLOT = wx.Window.NewControlId()
    PRINT_PLOT = wx.Window.NewControlId()
    PRINT_GRID = wx.Window.NewControlId()
    PRINT_SCRIPT = wx.Window.NewControlId()

    QUIT = wx.Window.NewControlId()

    UNDO = wx.Window.NewControlId()
    REDO = wx.Window.NewControlId()
    HISTORY = wx.Window.NewControlId()
    COPY_GRAPH = wx.Window.NewControlId()
    COPY_SIM = wx.Window.NewControlId()
    COPY_TABLE = wx.Window.NewControlId()
    FIND_REPLACE = wx.Window.NewControlId()
    OPEN_IN_EDITOR = wx.Window.NewControlId()

    NEW_DATA = wx.Window.NewControlId()
    DELETE_DATA = wx.Window.NewControlId()
    LOWER_DATA = wx.Window.NewControlId()
    RAISE_DATA = wx.Window.NewControlId()
    TOGGLE_SHOW = wx.Window.NewControlId()
    TOGGLE_USE = wx.Window.NewControlId()
    TOGGLE_ERROR = wx.Window.NewControlId()
    CALCS_DATA = wx.Window.NewControlId()

    TOGGLE_SLIDER = wx.Window.NewControlId()
    ZOOM = wx.Window.NewControlId()
    ZOOM_ALL = wx.Window.NewControlId()
    Y_SCALE_LIN = wx.Window.NewControlId()
    Y_SCALE_LOG = wx.Window.NewControlId()
    X_SCALE_LIN = wx.Window.NewControlId()
    X_SCALE_LOG = wx.Window.NewControlId()
    AUTO_SCALE = wx.Window.NewControlId()
    USE_TOGGLE_SHOW = wx.Window.NewControlId()

    SIM_MODEL = wx.Window.NewControlId()
    EVAL_MODEL = wx.Window.NewControlId()
    TOGGLE_CUDA = wx.Window.NewControlId()
    START_FIT = wx.Window.NewControlId()
    STOP_FIT = wx.Window.NewControlId()
    RESTART_FIT = wx.Window.NewControlId()
    CALC_ERROR = wx.Window.NewControlId()
    ANALYZE = wx.Window.NewControlId()
    AUTO_SIM = wx.Window.NewControlId()

    SET_OPTIMIZER = wx.Window.NewControlId()
    SET_DATA_LOADER = wx.Window.NewControlId()
    SET_IMPORT = wx.Window.NewControlId()
    SET_PLOT = wx.Window.NewControlId()
    SET_PROFILE = wx.Window.NewControlId()
    SET_EDITOR = wx.Window.NewControlId()
    SET_SINGLE = wx.Window.NewControlId()

    HELP_MODEL = wx.Window.NewControlId()
    HELP_FOM = wx.Window.NewControlId()
    HELP_PLUGINS = wx.Window.NewControlId()
    HELP_DATA_LOADERS = wx.Window.NewControlId()
    HELP_MANUAL = wx.Window.NewControlId()
    HELP_EXAMPLES = wx.Window.NewControlId()
    HELP_HOMEPAGE = wx.Window.NewControlId()
    HELP_ABOUT = wx.Window.NewControlId()
    HELP_DEBUG = wx.Window.NewControlId()


def rebuild_IDs():
    # re-generate above Enums with new IDs for e.g. window recreation
    global ToolId, MenuId
    new_vals = dict([(key, wx.Window.NewControlId()) for key in ToolId.__members__.keys()])
    ToolId = Enum("ToolId", new_vals, type=int, module=__name__)
    new_vals = dict([(key, wx.Window.NewControlId()) for key in MenuId.__members__.keys()])
    MenuId = Enum("MenuId", new_vals, type=int, module=__name__)



"""
Module for usefull decorators e.g. for logging function calls, input and output.
"""

import inspect
import logging
import threading

from functools import update_wrapper
from io import StringIO
from time import time


def getinfo(func):
    """
    Returns an info dictionary containing:
    - name (the name of the function : str)
    - argnames (the names of the arguments : list)
    - defaults (the values of the default arguments : tuple)
    - signature (the signature : str)
    - doc (the docstring : str)
    - module (the module name : str)
    - dict (the function __dict__ : str)
    """
    assert inspect.ismethod(func) or inspect.isfunction(func)
    signature = inspect.signature(func)
    output = dict(
        name=func.__name__,
        argnames=list(signature.parameters.keys()),
        signature=str(signature).lstrip("(").rstrip(")"),
        defaults=func.__defaults__,
        doc=func.__doc__,
        module=func.__module__,
        dict=func.__dict__,
        globals=func.__globals__,
        closure=func.__closure__,
    )
    return output


# the real meat is here
def _decorator(caller, func):
    infodict = getinfo(func)
    argnames = infodict["argnames"]
    assert not ("_call_" in argnames or "_func_" in argnames), "You cannot use _call_ or _func_ as argument names!"
    src = "lambda %(signature)s: _call_(_func_, %(signature)s)" % infodict
    dec_func = eval(src, dict(_func_=func, _call_=caller))
    return update_wrapper(dec_func, func)


def decorator(caller):
    """
    General purpose decorator factory: takes a caller function as
    input and returns a decorator with the same attributes.
    """
    return update_wrapper(lambda f: _decorator(caller, f), caller)


########################### Decorators for logging ##################################


def _nicetype(item):
    if "array" in type(item).__name__:
        return type(item).__name__ + ("[%s]" % str(item.shape))
    elif type(item) in [list, tuple]:
        return type(item).__name__ + ("[%i]" % len(item))
    else:
        return type(item).__name__


class DecoratorLogger(logging.getLoggerClass()):
    """
    A logger that makes sure the actual function definition filename, lineno and function name
    is used for logging.
    """

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
        if extra is None:
            return logging.getLoggerClass().makeRecord(
                self, name, level, fn, lno, msg, args, exc_info, func=func, extra=None
            )
        else:
            return logging.getLoggerClass().makeRecord(
                self, name, level, extra["name"], extra["lno"], msg, args, exc_info, func=extra["func"], extra=None
            )


# only use the debug decorators in debugging runs.
old_class = logging.getLoggerClass()
logging.setLoggerClass(DecoratorLogger)
logger = logging.getLogger("deco")
logger.setLevel(logging.DEBUG)
null_handler = logging.StreamHandler(StringIO())
null_handler.setLevel(logging.CRITICAL)
logger.addHandler(null_handler)
logging.setLoggerClass(old_class)


def _logformat(msg, decname, func):
    if hasattr(func, "im_func"):
        co = func.__func__.__code__
    else:
        co = func.__code__
    fname = co.co_filename
    lno = co.co_firstlineno
    logger.debug(msg, extra={"func": "@" + decname, "name": fname, "lno": lno})


@decorator
def log_call(func, *args, **kw):
    """
    Decorator to log just the method call.
    """
    if logging.root.getEffectiveLevel() > logging.DEBUG:
        return func(*args, **kw)
    infodict = getinfo(func)
    if len(infodict["argnames"]) > 0 and infodict["argnames"][0] == "self":
        _logformat(
            "%s.%s.%s <%s>"
            % (infodict["module"], args[0].__class__.__name__, infodict["name"], repr(threading.current_thread())),
            "log_call",
            func,
        )
    else:
        _logformat(
            "%s.%s <%s>" % (infodict["module"], infodict["name"], repr(threading.current_thread())), "log_call", func
        )
    return func(*args, **kw)


@decorator
def log_input(func, *args, **kw):
    """
    Decorator to log a method call with input.
    """
    if logging.root.getEffectiveLevel() > logging.DEBUG:
        return func(*args, **kw)
    infodict = getinfo(func)
    if hasattr(func, "im_func"):
        logstr = " call %s.%s(" % (args[0].__class__.__name__, infodict["name"])
        method = True
        info_len = len("%s.%s(" % (args[0].__class__.__name__, infodict["name"]))
    else:
        logstr = "call %s(" % (infodict["name"])
        method = False
        info_len = len("%s(" % (infodict["name"]))

    namelen = max([len(name) + 3 for name in infodict["argnames"]] + [len(name) for name in list(kw.keys())])
    for i, arg in enumerate(args):
        if i == 0 and (method or infodict["name"] == "__init__"):
            continue
        value = repr(arg)
        value_split = value.splitlines()
        if len(value_split) > 5:
            value = "\n".join(value_split[:2] + [" " * int(len(value_split[1]) / 2) + "..."] + value_split[-2:])
        value.replace("\n", "\n" + " " * int(namelen + 7))
        logstr += "\n" + ("%% %is= %%s (%%s)" % (namelen + 3)) % (infodict["argnames"][i], value, _nicetype(arg))

    for key, arg in sorted(kw.items()):
        value = repr(arg)
        value_split = value.splitlines()
        if len(value_split) > 5:
            value = "\n".join(value_split[:2] + [" " * int(len(value_split[1]) / 2) + "..."] + value_split[-2:])
        value.replace("\n", "\n" + " " * int(namelen + 4))
        logstr += "\n" + ("KW %% %ss= %%s (%%s)" % namelen) % (key, repr(value), _nicetype(arg))
    logstr += "\n)"
    logstr = logstr.replace("\n", "\n" + " " * int(info_len + 10))
    logstr = logstr.replace("call ", "call \n" + " " * 10)

    _logformat(logstr, "log_input", func)
    return func(*args, **kw)


@decorator
def log_output(func, *args, **kw):
    """
    Decorator to log a method call with output. If combined with log_input
    the input is logged at the time before the call and the output after.
    """
    if logging.root.getEffectiveLevel() > logging.DEBUG:
        return func(*args, **kw)
    output = func(*args, **kw)
    infodict = getinfo(func)
    if len(infodict["argnames"]) > 0 and infodict["argnames"][0] == "self":
        logstr = "return from %s.%s" % (args[0].__class__.__name__, infodict["name"])
    else:
        logstr = "return from %s" % (infodict["name"])
    value = repr(output)
    value_split = value.splitlines()
    if len(value_split) > 5:
        value = "\n".join(value_split[:2] + [" " * int(len(value_split[1]) / 2) + "..."] + value_split[-2:])
    logstr += "\n-> %15s (%s)" % (value, _nicetype(output))
    logstr = logstr.replace("\n", "\n" + " " * 44)
    _logformat(logstr, "log_output", func)
    return output


@decorator
def log_both(func, *args, **kw):
    """
    Decoratore to log a method call with input and output.
    """
    if logging.root.getEffectiveLevel() > logging.DEBUG:
        return func(*args, **kw)
    infodict = getinfo(func)
    if hasattr(func, "im_func"):
        logstr = " call %s.%s(" % (args[0].__class__.__name__, infodict["name"])
        method = True
        info_len = len("%s.%s(" % (args[0].__class__.__name__, infodict["name"]))
    else:
        logstr = "call %s(" % (infodict["name"])
        method = False
        info_len = len("%s(" % (infodict["name"]))

    namelen = max([len(name) + 3 for name in infodict["argnames"]] + [len(name) for name in list(kw.keys())])
    for i, arg in enumerate(args):
        if i == 0 and (method or infodict["name"] == "__init__"):
            continue
        value = repr(arg)
        value_split = value.splitlines()
        if len(value_split) > 5:
            value = "\n".join(value_split[:2] + [" " * int(len(value_split[1]) / 2) + "..."] + value_split[-2:])
        value.replace("\n", "\n" + " " * int(namelen + 7))
        logstr += "\n" + ("%% %is= %%s (%%s)" % (namelen + 3)) % (infodict["argnames"][i], value, _nicetype(arg))

    for key, arg in sorted(kw.items()):
        value = repr(arg)
        value_split = value.splitlines()
        if len(value_split) > 5:
            value = "\n".join(value_split[:2] + [" " * int(len(value_split[1]) / 2) + "..."] + value_split[-2:])
        value.replace("\n", "\n" + " " * int(namelen + 4))
        logstr += "\n" + ("KW %% %ss= %%s (%%s)" % namelen) % (key, repr(value), _nicetype(arg))
    logstr += "\n)"
    logstr = logstr.replace("\n", "\n" + " " * int(info_len + 10))
    logstr = logstr.replace("call ", "call \n" + " " * 10)

    _logformat(logstr, "log_input", func)
    # call the function
    output = func(*args, **kw)
    if len(infodict["argnames"]) > 0 and infodict["argnames"][0] == "self":
        logstr = "return from %s.%s" % (args[0].__class__.__name__, infodict["name"])
    else:
        logstr = "return from %s" % (infodict["name"])
    value = repr(output)
    value_split = value.splitlines()
    if len(value_split) > 5:
        value = "\n".join(value_split[:2] + [" " * int(len(value_split[1]) / 2) + "..."] + value_split[-2:])
    logstr += "\n-> %s (%s)" % (value, _nicetype(output))
    logstr = logstr.replace("\n", "\n" + " " * int(info_len + 10))
    _logformat(logstr, "log_output", func)
    return output


timings = {}


@decorator
def time_call(func, *args, **kw):
    """
    Decorator to log just the method call.
    """
    if logging.root.getEffectiveLevel() > logging.DEBUG:
        return func(*args, **kw)
    name = func.__name__
    start = time()
    output = func(*args, **kw)
    runtime = time() - start
    if not name in timings:
        timings[name] = (0.0, 0.0)
    avg, calls = timings[name]
    timings[name] = ((avg * calls + runtime) / (calls + 1), calls + 1)
    return output


########################## General decorators ###############################


class check_input(object):
    """
    Decorator checking the input to a function.

    Usage:
    @check_input([type1, type2, ...], try_convert=True)
    def your_function(var1, var2, ...):
        ... function text ...
    """

    try_convert = True
    types = []

    def __init__(self, types, try_convert=True):
        self.types = types
        self.try_convert = try_convert

    def __call__(self, func):
        infodict = getinfo(func)
        argnames = infodict["argnames"]
        assert not ("_call_" in argnames or "_func_" in argnames), "You cannot use _call_ or _func_ as argument names!"
        src = "def dec_func(%(signature)s):" % infodict
        for i, typei in enumerate(self.types):
            src += '\n  if type(%s).__name__!="%s":' % (argnames[i], typei.__name__)
            if self.try_convert:
                src += "\n    try:"
                src += "\n      %s=%s(%s)" % (argnames[i], typei.__name__, argnames[i])
                src += "\n    except:"
                src += '\n      raise ValueError("type of %s is not %s")' % (argnames[i], typei.__name__)
            else:
                src += '\n    raise ValueError("type of %s is not %s")' % (argnames[i], typei.__name__)
        src += "\n  return _func_(%(signature)s)" % infodict
        exec_dict = dict(_func_=func, _call_=self.__call__)
        exec(src, exec_dict)
        return update_wrapper(exec_dict["dec_func"], func)



"""
======
MagSLD
======

A plugin to show a second y-axis in the SLD plot that gives the magnetization corresponging
to the magnetic SLD and to calculate the integrated magnetic moment surface density. 

Written by Artur Glavic
Last Changes 04/28/15
"""

import wx

from numpy import trapz

from genx.models.lib.physical_constants import AAm2_to_emucc

from .. import add_on_framework as framework


class Plugin(framework.Template):
    def __init__(self, parent):
        framework.Template.__init__(self, parent)
        self.parent = parent

        self._init_refplugin()

        menu = self.NewMenu("Mag. SLD")
        self.mb_second_axis = wx.MenuItem(
            menu, wx.NewId(), "Show second SLD axis", "Show second SLD axis for magnetization", wx.ITEM_CHECK
        )
        menu.Append(self.mb_second_axis)
        self.mb_second_axis.Check(True)
        self.mb_use_SI = wx.MenuItem(menu, wx.NewId(), "Use SI units", "Use SI units of magnetization", wx.ITEM_CHECK)
        menu.Append(self.mb_use_SI)
        self.mb_use_SI.Check(False)

        self.parent.Bind(wx.EVT_MENU, self._OnSimulate, self.mb_second_axis)
        self.parent.Bind(wx.EVT_MENU, self._OnSimulate, self.mb_use_SI)

    def _init_refplugin(self):
        ph = self.parent.plugin_control.plugin_handler
        if "Reflectivity" in ph.loaded_plugins:
            # connect to the reflectivity plugin for layer creation
            self.refplugin = ph.loaded_plugins["Reflectivity"]
        elif "SimpleReflectivity" in ph.loaded_plugins:
            # connect to the reflectivity plugin for layer creation
            self.refplugin = ph.loaded_plugins["SimpleReflectivity"]
        else:
            raise RuntimeError("Can't load plugin without Reflectivity plugin present.")

        self._orgi_call = self.refplugin.OnSimulate
        self.refplugin.OnSimulate = self._OnSimulate

        sld_plot = self.refplugin.sld_plot
        ax = sld_plot.ax
        self._orig_position = ax.get_position()
        ax2 = ax.twinx()
        sld_plot.ax2 = ax2
        box = self._orig_position
        ax2.set_position([box.x0, box.y0, box.width, box.height])
        ax2.set_visible(False)
        self._annotations = []

    def Remove(self):
        self.refplugin.OnSimulate = self._orgi_call
        sld_plot = self.refplugin.sld_plot
        sld_plot.ax2.set_visible(False)
        del sld_plot.ax2
        del self.refplugin
        framework.Template.Remove(self)

    def _OnSimulate(self, event):
        """
        Call original reflectivity plugin OnSimulate function to make sure the plot is
        already created when adding the second axis.
        """
        for ann in self._annotations:
            try:
                ann.remove()
            except Exception:
                pass
        self._annotations = []
        self._orgi_call(event)
        sld_plot = self.refplugin.sld_plot
        model = self.GetModel()
        sld_items = len(sld_plot.plot_dicts)
        msld = None
        z = None
        unit = ""
        ax = sld_plot.ax
        ax2 = sld_plot.ax2
        for i, di in enumerate(model.data):
            if di.show and sld_items > i:
                slds = sld_plot.plot_dicts[i]
                if not "z" in slds or not "SLD unit" in slds:
                    continue
                z = slds["z"]
                unit = slds["SLD unit"]
                if unit in ["fm/\\AA^{3}", "\\AA^{-2}", "", "10^{-6}\\AA^{-2}"] and self.mb_second_axis.IsChecked():
                    for key, value in list(slds.items()):
                        if key == "mag" and value.sum() > 0.0:
                            msld = value
                            mag = msld * AAm2_to_emucc
                            com = (msld * z).sum() / msld.sum()
                            if self.mb_use_SI.IsChecked():
                                self._annotations.append(
                                    ax.annotate(
                                        "Integrated M:\n%.4g $A$" % trapz(mag * 1e3, z * 1e-10),
                                        (com, 0.02),
                                        ha="center",
                                    )
                                )
                            else:
                                self._annotations.append(
                                    ax.annotate(
                                        "Integrated M:\n%.4g $emu/cm^2$" % trapz(mag, z * 1e-8),
                                        (com, 0.02),
                                        ha="center",
                                    )
                                )
        if (
            msld is not None
            and unit in ["fm/\\AA^{3}", "\\AA^{-2}", "", "10^{-6}\\AA^{-2}"]
            and self.mb_second_axis.IsChecked()
        ):
            ax2.set_visible(True)
            ymin, ymax = ax.get_ylim()
            if self.mb_use_SI.IsChecked():
                ax2.set_ylim((ymin * AAm2_to_emucc * 1e3, ymax * AAm2_to_emucc * 1e3))
                ax2.yaxis.label.set_text("M [$A/m$]")
            else:
                ax2.set_ylim((ymin * AAm2_to_emucc, ymax * AAm2_to_emucc))
                ax2.yaxis.label.set_text("M [$emu/cm^3$]")
            sld_plot.ax.legend(loc="upper right", framealpha=0.5, fontsize="small", ncol=1)  # bbox_to_anchor=(1, 0.5),
        else:
            ax2.set_visible(False)



"""
A general wx.Dialog to deal with user configurable settings.
"""

from math import log10

import wx

from ..core.config import BaseConfig


class SettingsDialog(wx.Dialog):
    def __init__(self, parent: wx.Window, settings: BaseConfig, title="Settings", apply_callback: callable = None):
        """
        Configuration dialog for BaseConfig based classes.

        The apply_callback is a function that is called with a dictionary of updated options that
        can be evaluated. It returns a bool to determin if the dialog should apply these options to the
        config object.
        """
        wx.Dialog.__init__(self, parent, -1, title, style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        self.settings = settings
        if apply_callback is None:
            self.apply_callback = lambda options: True
        else:
            self.apply_callback = apply_callback

        self.build_sizers()
        self.rb_actions = {}  # entries for toggling radio button selected entries enabled
        self.rb_bool_fields = {}  # the associated with the above action
        self.value_entries = {}  # all controls for settable values, the key is field.name

        groups = list(self.settings.groups.keys())
        n_groups = len(groups)
        r_groups = n_groups // 2
        l_groups = n_groups - r_groups
        for lg in groups[:l_groups]:
            self.build_group(lg, self.row_sizer1)
        for rg in groups[l_groups:]:
            self.build_group(rg, self.row_sizer2)

        self.GetSizer().Fit(self)
        self.Layout()

    def build_sizers(self, add_apply=False):
        col_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.row_sizer1 = wx.BoxSizer(wx.VERTICAL)
        self.row_sizer2 = wx.BoxSizer(wx.VERTICAL)
        col_sizer.Add(self.row_sizer1, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)
        col_sizer.Add(self.row_sizer2, 0, wx.ALIGN_CENTER_VERTICAL | wx.ALL, 5)

        # Add the Dialog buttons
        button_sizer = wx.StdDialogButtonSizer()
        okay_button = wx.Button(self, wx.ID_OK)
        okay_button.SetDefault()
        button_sizer.AddButton(okay_button)
        if add_apply:
            apply_button = wx.Button(self, wx.ID_APPLY)
            apply_button.SetDefault()
            button_sizer.AddButton(apply_button)
        button_sizer.AddButton(wx.Button(self, wx.ID_CANCEL))
        button_sizer.Realize()
        # Add some event handlers
        self.Bind(wx.EVT_BUTTON, self.on_apply_changes)

        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(col_sizer, 1, wx.GROW | wx.LEFT | wx.RIGHT, 10)
        line = wx.StaticLine(self, -1, size=(20, -1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW | wx.TOP, 10)

        sizer.Add(button_sizer, 0, flag=wx.ALIGN_RIGHT, border=4)
        self.SetSizer(sizer)

    def build_group(self, name: str, sizer: wx.BoxSizer):
        box = wx.StaticBox(self, -1, name)
        box_sizer = wx.StaticBoxSizer(box, orient=wx.VERTICAL)
        sizer.Add(box_sizer, 0, wx.EXPAND)

        pnames = self.settings.groups[name]
        fields = dict((fi.name, fi) for fi in self.settings.get_fields())
        for ni in pnames:
            if type(ni) in [list, tuple]:
                if len(ni) == 4 and fields[ni[1]].type is bool:
                    # the entry is toggling type
                    e1_active = fields[ni[1]]
                    e1_field = fields[ni[2]]
                    e2_field = fields[ni[3]]
                    gbox = wx.StaticBox(box, -1, ni[0])
                    gbox_sizer = wx.StaticBoxSizer(gbox, orient=wx.VERTICAL)
                    box_sizer.Add(gbox_sizer, 0, wx.EXPAND)
                    self.rb_bool_fields[gbox.GetId()] = ni[1]
                    e1 = self.get_toggle_entry(e1_field, gbox, getattr(self.settings, e1_active.name, True))
                    gbox_sizer.Add(e1, 1, wx.EXPAND | wx.ALL, 2)
                    e2 = self.get_toggle_entry(e2_field, gbox, not getattr(self.settings, e1_active.name, True))
                    gbox_sizer.Add(e2, 1, wx.EXPAND | wx.BOTTOM, 5)
                    gbox.Bind(wx.EVT_RADIOBUTTON, self.on_radio)
                else:
                    hsizer = wx.BoxSizer(wx.HORIZONTAL)
                    box_sizer.Add(hsizer, 0, wx.ALIGN_CENTRE | wx.ALL, 5)
                    for nj in ni:
                        field = fields[nj]
                        entry = self.get_entry(field, box)
                        hsizer.Add(entry, 0, wx.EXPAND | wx.ALL, 2)
            else:
                field = fields[ni]
                entry = self.get_entry(field, box)
                box_sizer.Add(entry, 1, wx.FIXED_MINSIZE | wx.ALIGN_CENTRE | wx.ALL, 5)
        box_sizer.Fit(box)

    def get_entry(self, field, parent, add_label=True):
        gmeta = field.metadata.get("genx", {})
        if "label" in gmeta:
            name = gmeta["label"]
        else:
            name = field.name.replace("_", " ")
        if field.type is bool:
            entry = wx.CheckBox(parent, -1, name, style=wx.ALIGN_CENTER)
            entry.SetValue(getattr(self.settings, field.name, False))
            self.value_entries[field.name] = entry
        elif "selection" in gmeta:
            choice = wx.Choice(parent, -1, choices=gmeta["selection"])
            choice.SetSelection(gmeta["selection"].index(getattr(self.settings, field.name, gmeta["selection"][0])))
            self.value_entries[field.name] = choice
            entry = self.label_entry(parent, choice, name)
        elif "pmin" in gmeta and "pmax" in gmeta:
            # a numerical parameter with given range
            if field.type is int:
                spin = wx.SpinCtrl(parent, -1, min=gmeta["pmin"], max=gmeta["pmax"])
            else:
                inc = 10 ** (round(log10(gmeta["pmax"])) - 2)
                spin = wx.SpinCtrlDouble(parent, -1, min=gmeta["pmin"], max=gmeta["pmax"], inc=inc)
                spin.SetDigits(4)
            spin.SetValue(getattr(self.settings, field.name, gmeta["pmin"]))

            spin.SetMinSize(spin.GetSizeFromTextSize(spin.GetTextExtent("%.4f" % gmeta["pmax"] + " ")))
            self.value_entries[field.name] = spin
            if add_label:
                entry = self.label_entry(parent, spin, name)
            else:
                entry = spin
        else:
            txt = wx.TextCtrl(parent, -1, style=0)
            txt.SetValue(str(getattr(self.settings, field.name, "")))
            txt.SetMinSize(txt.GetSizeFromTextSize(txt.GetTextExtent(10 * " ")))
            self.value_entries[field.name] = txt
            if add_label:
                entry = self.label_entry(parent, txt, name)
            else:
                entry = txt
        return entry

    def get_toggle_entry(self, field, parent, active):
        gmeta = field.metadata.get("genx", {})
        if "label" in gmeta:
            name = gmeta["label"]
        else:
            name = field.name.replace("_", " ")
        # generate an entries with a radio button to toggle it active
        entry = wx.BoxSizer(wx.HORIZONTAL)
        entry_active = wx.RadioButton(parent, -1, name, style=wx.ALIGN_RIGHT)
        entry_active.SetValue(active)
        entry.Add(entry_active, 1, wx.FIXED_MINSIZE | wx.RIGHT, 8)
        field_entry = self.get_entry(field, parent, add_label=False)
        field_entry.Enable(active)
        parent_id = parent.GetId()
        if parent_id in self.rb_actions:
            self.rb_actions[parent_id][entry_active.GetId()] = field_entry
        else:
            self.rb_actions[parent_id] = {entry_active.GetId(): field_entry}
        entry.Add(field_entry, 1, wx.FIXED_MINSIZE, 0)
        return entry

    def on_radio(self, event: wx.CommandEvent):
        # activates the entry next to the selected radio button and deactivates all others in that group
        parent_id = event.GetEventObject().GetParent().GetId()
        sender_id = event.GetId()
        for id, field_entry in self.rb_actions[parent_id].items():
            field_entry.Enable(sender_id == id)
        event.Skip()

    def label_entry(self, parent, entry, label):
        out_entry = wx.BoxSizer(wx.HORIZONTAL)
        out_entry.Add(
            wx.StaticText(parent, -1, label, style=wx.ALIGN_RIGHT | wx.ALIGN_CENTER_VERTICAL),
            0,
            wx.FIXED_MINSIZE | wx.ALIGN_CENTER_VERTICAL | wx.RIGHT,
            8,
        )
        out_entry.Add(entry, 0, wx.EXPAND, 0)
        return out_entry

    def collect_results(self):
        # create a dictionary with all values from the dialog entry
        out = {}
        fields = dict((fi.name, fi) for fi in self.settings.get_fields())
        for key, entry in self.value_entries.items():
            field = fields[key]
            if hasattr(entry, "GetValue"):
                try:
                    out[key] = field.type(entry.GetValue())
                except ValueError:
                    # TODO: add handling of entry errors
                    pass
            elif hasattr(entry, "GetString") and hasattr(entry, "GetSelection"):
                out[key] = entry.GetString(entry.GetSelection())
            else:
                raise NotImplementedError(f"Could not evaluate wx Control {entry!r}")
        for pid, rbg in self.rb_actions.items():
            # get the active state of the boolean fields for radio buttons
            out[self.rb_bool_fields[pid]] = list(rbg.values())[0].IsEnabled()
        return out

    def on_apply_changes(self, event):
        event.Skip()
        options = self.collect_results()
        OK_pressed = event.GetId() == wx.ID_OK
        if self.apply_callback(options) and OK_pressed:
            for key, value in options.items():
                setattr(self.settings, key, value)



import string

import wx


class SliderControl(wx.Slider):
    """A custom slider control."""

    def __init__(
        self, parent, id, value=0.0, min_value=0.0, max_value=100.0, font=None, border_frame=True, size=wx.DefaultSize
    ):
        int_value = int(100 * (value - min_value) / (max_value - min_value))
        wx.Slider.__init__(self, parent, id, value=int_value, minValue=0, maxValue=100, size=size)
        self.min_value = min_value
        self.max_value = max_value
        self.value = float(value)

        self.border_frame = border_frame
        self.scroll_callback = None

    def bind_handlers(self):
        self.Bind(wx.EVT_CHAR, self.OnChar)
        self.Bind(wx.EVT_SLIDER, self.OnChange)

    def unbind_handlers(self):
        self.Unbind(wx.EVT_SLIDER)

    def OnChar(self, event):
        if chr(event.GetKeyCode()) == " ":
            self.GetParent().GetParent().SaveEditControlValue()
            self.GetParent().GetParent().EnableCellEditControl(False)
            return
        else:
            event.Skip()

    def OnChange(self, event):
        self.value = self.GetValue()

        if self.scroll_callback is not None:
            self.scroll_callback(self.value)

    def set_slider_position(self):
        if self.min_value == self.max_value:
            int_value = 50
        else:
            int_value = int(100 * (self.value - self.min_value) / (self.max_value - self.min_value))
        wx.Slider.SetValue(self, int_value)

    def SetValue(self, value):
        self.value = value
        self.set_slider_position()

    def GetValue(self):
        int_value = wx.Slider.GetValue(self)
        return int_value / 100.0 * (self.max_value - self.min_value) + self.min_value

    def SetMaxValue(self, value):
        self.max_value = value
        self.set_slider_position()

    def SetMinValue(self, value):
        self.min_value = value
        self.set_slider_position()

    def SetScrollCallback(self, func):
        """Set a callback that is called on a scroll event.

        func should only take one parameter which is the value, i.e. func(value).
        """
        self.scroll_callback = func


class NumberValidator(wx.Validator):
    """Validaator to handle numerical values, accepts also 'e' as exponent value as input"""

    def __init__(self):
        wx.Validator.__init__(self)
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def Clone(self):
        return NumberValidator()

    def Validate(self, win):
        tc = self.GetWindow()
        val = tc.GetValue()

        try:
            val = float(val)
        except:
            return False
        else:
            return True

    def OnChar(self, event):
        if event.AltDown() or event.ControlDown():
            event.Skip()
            return
        key = event.GetKeyCode()
        tc = self.GetWindow()
        val = tc.GetValue()
        pos = tc.GetInsertionPoint()

        if event.GetUnicodeKey() != wx.WXK_NONE:
            # an actual character key was pressed

            if chr(key) in string.digits:
                if len(val) == 0:
                    event.Skip()
                    return
                if pos == 0 and val[0] == "-":
                    if not wx.Validator_IsSilent():
                        wx.Bell()
                    return
                event.Skip()
                return

            if chr(key) == "." and "." not in val:
                if "e" in val:
                    # Check so that the . ends up correctly relative e
                    if pos > val.index("e"):
                        if not wx.Validator_IsSilent():
                            wx.Bell()
                        return
                if len(val) == 0:
                    event.Skip()
                    return
                if pos == 0 and val[0] == "-":
                    if not wx.Validator_IsSilent():
                        wx.Bell()
                    return

                event.Skip()
                return

            if chr(key) == "e" and "e" not in val:
                if "." in val:
                    # Check so that the e ends up correctly relative .
                    if pos <= val.index("."):
                        if not wx.Validator_IsSilent():
                            wx.Bell()
                        return
                if len(val) == 0:
                    event.Skip()
                    return
                if pos == 0 and val[0] == "-":
                    if not wx.Validator_IsSilent():
                        wx.Bell()
                    return
                event.Skip()
                return

            if chr(key) == "-":
                if val[0] != "-" and pos == 0:
                    event.Skip()
                elif val[pos - 1] == "e" and "-" not in val[pos - 1 :]:
                    event.Skip()
                elif val[0] == "-":
                    tc.SetValue(val[1:])
                    tc.SetInsertionPoint(max(0, pos - 1))
                else:
                    tc.SetValue("-" + val)
                    tc.SetInsertionPoint(pos + 1)
                return

            if chr(key) == " " and self.Validate(None):
                tc.GetParent().GetParent().SaveEditControlValue()
                tc.GetParent().GetParent().EnableCellEditControl(False)
                return

        if key in [
            wx.WXK_TAB,
            wx.WXK_RETURN,
            wx.WXK_BACK,
            wx.WXK_ESCAPE,
            wx.WXK_DOWN,
            wx.WXK_UP,
            wx.WXK_LEFT,
            wx.WXK_RIGHT,
            wx.WXK_NUMPAD_LEFT,
            wx.WXK_NUMPAD_RIGHT,
            wx.WXK_DELETE,
            wx.WXK_BACK,
            wx.WXK_HOME,
            wx.WXK_END,
        ]:
            event.Skip()
            return

        if not wx.Validator.IsSilent():
            wx.Bell()

        # Does not allow the event to propagate (kills it)
        return


class SpinCtrl(wx.TextCtrl):
    def __init__(
        self,
        parent,
        id=wx.NewId(),
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        value=0.0,
        min_value=None,
        max_value=None,
        steps=20,
        digits=6,
        name="SpinCtrl",
    ):
        self.value = value
        self.digits = digits
        self.min_value = min_value
        self.max_value = max_value
        self.steps = steps
        self.value_change_func = None
        wx.TextCtrl.__init__(
            self,
            parent,
            id=id,
            pos=pos,
            size=size,
            name=name,
            style=wx.TE_RIGHT,
            value="%.*g" % (self.digits, self.value),
            validator=NumberValidator(),
        )

        self.Layout()
        # self._spin=wx.SpinButton(self, wx.NewId(), pos=pos, size=wx.Size(20, self.GetSize().GetHeight()-8),
        #                          style=wx.SP_ARROW_KEYS | wx.SP_VERTICAL, name=name+"_SpinButton")
        # self._spin.Bind(wx.EVT_SPIN_UP, self.eh_spin_up,
        #                 id=self._spin.GetId())
        # self._spin.Bind(wx.EVT_SPIN_DOWN, self.eh_spin_down,
        #                 id=self._spin.GetId())

        self.Bind(wx.EVT_MOUSEWHEEL, self.OnMouseWheel)
        self.Bind(wx.EVT_CHAR, self.OnChar)

    def step(self, dir_up=True):
        """Steps the value up/down on increment

        :param dir_up: stepping direction
        :return:
        """
        inc = (self.max_value - self.min_value) * 1.0 / self.steps
        val = self.value
        if dir_up:
            val += inc
        else:
            val -= inc

        # Check so that the bounds are kept
        val = max(val, self.min_value)
        val = min(val, self.max_value)

        self.SetValue(val)

        if self.value_change_func is not None:
            self.value_change_func(val)

    def eh_spin_up(self, event):
        self.step(dir_up=True)
        if self.value_change_func is not None:
            self.value_change_func(self.GetValue())
        event.Skip()

    def eh_spin_down(self, event):
        self.step(dir_up=False)
        if self.value_change_func is not None:
            self.value_change_func(self.GetValue())
        event.Skip()

    def OnMouseWheel(self, event):
        if event.GetWheelRotation() > 0 and not event.IsWheelInverted():
            self.step(dir_up=True)
        else:
            self.step(dir_up=False)

    def OnChar(self, event):
        # print event.GetKeyCode()
        if event.GetKeyCode() == wx.WXK_UP:
            self.step(dir_up=True)
        elif event.GetKeyCode() == wx.WXK_DOWN:
            self.step(dir_up=False)
        else:
            event.Skip()

    def SetValue(self, value):
        try:
            self.value = float(value)
            wx.TextCtrl.SetValue(self, "%.*g" % (self.digits, self.value))
        except ValueError:
            wx.TextCtrl.SetValue(self, value)

    def GetValue(self):
        try:
            self.value = float(wx.TextCtrl.GetValue(self))
        except ValueError:
            pass
        return wx.TextCtrl.GetValue(self)

    def SetValueChangeCallback(self, func):
        """Sets a callback function to execute when the value has changed
        :param func:
        :return:
        """
        self.value_change_func = func

    def SetRange(self, min_value, max_value):
        """Sets the range of the control

        :param min_value:
        :param max_value:
        :return:
        """
        if max_value < min_value:
            tmp_value = min_value
            min_value = max_value
            max_value = tmp_value
        self.min_value = float(min_value)
        self.max_value = float(max_value)



__version__ = "3.8.8"

def increment_version(minor=True, current=None):
    if current is None:
        current = __version__
    import os
    location = os.path.abspath(__file__)
    module_lines = open(location, 'r').readlines()

    full, ver, vminor = map(int, current.split('.'))
    if minor:
        vminor += 1
    else:
        ver += 1
        vminor = 0
    new_version = f"{full}.{ver}.{vminor}"
    for i, line in enumerate(module_lines):
        if line.startswith("__version__"):
            module_lines[i] = f'__version__ = "{new_version}"\n'
    with open(location, 'w') as f:
        f.writelines(module_lines)



"""
Classes to interface with the bumps module for fitting and statistical analysis.
"""

import threading

import bumps
import wx

from bumps import __version__ as bumps_version
from bumps.dream.corrplot import _hists
if tuple(bumps_version.split('.'))<('1','0','3'):
    from bumps.fitproblem import nllf_scale
else:
    # Adapt to changed signature of nllf_scale function
    from bumps.fitproblem import nllf_scale as _nllf_scale
    def nllf_scale(problem, norm=True):
        dof = problem.dof if norm else 1
        npars = max(len(problem.getp()), 1)
        return _nllf_scale(dof, npars, norm=norm)
from bumps.formatnum import format_uncertainty
from bumps.monitor import TimedUpdate
from matplotlib.colors import LogNorm
from numpy import array, maximum, newaxis, sqrt, linspace
from wx.grid import EVT_GRID_CELL_LEFT_DCLICK

from ..bumps_optimizer import BumpsResult
from ..plugins.utils import ShowInfoDialog
from .exception_handling import CatchModelError
from .plotpanel import BasePlotConfig, PlotPanel


class ProgressMonitor(TimedUpdate):
    """
    Display fit progress on the console
    """

    def __init__(self, problem, pbar, ptxt, progress=0.25, improvement=5.0):
        TimedUpdate.__init__(self, progress=progress, improvement=improvement)
        self.problem = problem
        self.pbar = pbar  # wx.Gauge
        self.ptxt = ptxt  # wx.StaticText
        self.chis = []
        self.steps = []

    def show_progress(self, history):
        scale, err = nllf_scale(self.problem)
        chisq = format_uncertainty(scale * history.value[0], err)
        wx.CallAfter(self.ptxt.SetLabel, "step: %s/%s  cost: %s" % (history.step[0], self.pbar.GetRange(), chisq))
        wx.CallAfter(self.pbar.SetValue, min(history.step[0], self.pbar.GetRange()))
        self.steps.append(history.step[0])
        self.chis.append(scale * history.value[0])

    def show_improvement(self, history):
        return
        p = self.problem.getp()
        try:
            self.problem.setp(history.point[0])
            out = '<table width="50%"><tr>'
            out += "</tr>\n<tr>".join(
                [
                    "<td>%s</td><td>%s</td><td>%s</td>" % (pi.name, pi.value, pi.bounds)
                    for pi in self.problem._parameters
                ]
            )
            self.result_text.value = out + "</td></tr></table>"
        finally:
            self.problem.setp(p)
        with self.plot_out:
            fig = figure()
            plot(self.steps, self.chis)
            clear_output()
            display(fig)
            close()


class HeaderCopyGrid(wx.grid.Grid):
    """Grid calss that copies the column and row headers together with the data."""

    def __init__(self, *args, **opts):
        wx.grid.Grid.__init__(self, *args, **opts)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKey)

    def OnKey(self, event):
        # If Ctrl+C is pressed...
        if event.ControlDown() and event.GetKeyCode() == 67:
            self.copy()
        else:
            event.Skip()
            return

    def copy(self):
        output = ""
        for block in self.GetSelectedBlocks():
            top, left = block.GetTopLeft()
            bottom, right = block.GetBottomRight()
            output += "\t"
            for col in range(left, right + 1):
                name = self.GetColLabelValue(col).replace("\n", ".")
                output += f"{name}\t"
            output = output[:-1] + "\n"
            for row in range(top, bottom + 1):
                name = self.GetRowLabelValue(row).replace("\n", ".")
                output += f"{name}\t"
                for col in range(left, right + 1):
                    output += f"{self.GetCellValue(row, col)}\t"
                output = output[:-1] + "\n"
            output += "\n\n"

        clipboard = wx.TextDataObject()
        clipboard.SetText(output)
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(clipboard)
            wx.TheClipboard.Close()
        else:
            wx.MessageBox("Can't open the clipboard", "Error")


class StatisticsPanelConfig(BasePlotConfig):
    section = "statistics plot"


class StatisticalAnalysisDialog(wx.Dialog):
    rel_cov = None
    thread: threading.Thread
    _res: BumpsResult

    def __init__(self, parent, model, prev_result: BumpsResult = None):
        wx.Dialog.__init__(self, parent, style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER | wx.MAXIMIZE_BOX)
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)
        self.SetTitle("Statistical Analysis of Parameters")

        dpi_scale_factor = wx.GetApp().dpi_scale_factor

        self.ptxt = wx.StaticText(self, label="...")
        self.pbar = wx.Gauge(self, range=1000)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        vbox.Add(hbox, proportion=1, flag=wx.EXPAND)

        lpanel = wx.Panel(self)
        gpanel = wx.Panel(self)
        gbox = wx.BoxSizer(wx.VERTICAL)
        gpanel.SetSizer(gbox)
        self.grid = HeaderCopyGrid(gpanel)
        gbox.Add(wx.StaticText(gpanel, label="Estimated covariance matrix:"), proportion=0, flag=wx.FIXED_MINSIZE)
        gbox.Add(self.grid, proportion=1, flag=wx.EXPAND)
        hbox2 = wx.BoxSizer(wx.HORIZONTAL)
        gbox.Add(hbox2, proportion=0, flag=wx.FIXED_MINSIZE)
        self.normalize_checkbox = wx.CheckBox(gpanel, label="Normalize value (σ_ij/σ_i/σ_j)")
        self.normalize_checkbox.SetValue(True)
        self.normalize_checkbox.Bind(wx.EVT_CHECKBOX, self.OnToggleNormalize)
        hbox2.Add(self.normalize_checkbox, proportion=0, flag=wx.FIXED_MINSIZE)
        self.chicorrect_checkbox = wx.CheckBox(gpanel, label="Correct by sqrt(chi²)")
        self.chicorrect_checkbox.SetToolTip(
            "Renormalize the parameter errors with sqrt(chi²).\n"
            "This is helpful if you assume the model is correct and "
            "error bars are only a relative measure.\n"
            "The resulting scaled error bars would then lead to chi²=1."
        )
        self.chicorrect_checkbox.SetValue(False)
        self.chicorrect_checkbox.Bind(wx.EVT_CHECKBOX, self.OnToggleChi2)
        hbox2.Add(self.chicorrect_checkbox, proportion=0, flag=wx.FIXED_MINSIZE)

        nfparams = len([pi for pi in model.parameters if pi.fit])
        self.grid.CreateGrid(nfparams + 2, nfparams)
        self.grid.SetColLabelTextOrientation(wx.VERTICAL)
        self.grid.SetColLabelSize(int(dpi_scale_factor * 80))
        for i in range(nfparams):
            self.grid.SetRowSize(i + 2, int(dpi_scale_factor * 80))
            self.grid.SetColSize(i, int(dpi_scale_factor * 80))
            self.grid.SetColLabelValue(i, "%i" % i)
            self.grid.SetRowLabelValue(i + 2, "%i" % i)
        self.grid.SetRowSize(0, int(dpi_scale_factor * 80 / 2))
        self.grid.SetRowSize(1, int(dpi_scale_factor * 80 / 2))
        self.grid.DisableCellEditControl()
        self.grid.SetMinSize((int(dpi_scale_factor * 200), int(dpi_scale_factor * 200)))
        self.grid.Bind(EVT_GRID_CELL_LEFT_DCLICK, self.OnSelectCell)

        rpbox = wx.BoxSizer(wx.VERTICAL)
        rpanel = wx.Panel(self)
        rpanel.SetSizer(rpbox)

        self.fom_text = wx.StaticText(rpanel, label="FOM chi²/bars: -")
        font = wx.Font(wx.FontInfo(2.0 * self.fom_text.GetFont().GetPointSize()))
        self.fom_text.SetFont(font)
        rpbox.Add(self.fom_text, proportion=0, flag=wx.FIXED_MINSIZE | wx.EXPAND)

        plot_panel = PlotPanel(rpanel, config_class=StatisticsPanelConfig)
        plot_panel.SetMinSize((int(dpi_scale_factor * 200), int(dpi_scale_factor * 200)))
        self.plot_panel = plot_panel
        self.ax = plot_panel.figure.add_subplot(111)
        rpbox.Add(plot_panel, proportion=1, flag=wx.EXPAND)

        hbox.Add(lpanel, proportion=0, flag=wx.EXPAND | wx.ALIGN_TOP)
        hbox.Add(gpanel, proportion=1, flag=wx.EXPAND)
        hbox.Add(rpanel, proportion=1, flag=wx.EXPAND)

        lsizer = wx.GridSizer(vgap=1, hgap=2, cols=2)
        lpanel.SetSizer(lsizer)
        self.entries = {}
        for key, emin, emax, val in [
            ("pop", 1, 20 * nfparams, 2 * nfparams),
            ("samples", 1000, 10000000, 10000),
            ("burn", 0, 10000, 200),
        ]:
            lsizer.Add(wx.StaticText(lpanel, label="%s:" % key), flag=wx.FIXED_MINSIZE)
            self.entries[key] = wx.SpinCtrl(lpanel, wx.ID_ANY, min=emin, max=emax, value=str(val))
            lsizer.Add(self.entries[key], flag=wx.FIXED_MINSIZE)

        lsizer.AddStretchSpacer(10)
        self.run_button = wx.Button(self, label="Run Analysis...")
        vbox.Add(self.run_button)
        vbox.Add(self.ptxt, proportion=0, flag=wx.EXPAND)
        vbox.Add(self.pbar, proportion=0, flag=wx.EXPAND)

        self.Bind(wx.EVT_BUTTON, self.OnRunAnalysis, self.run_button)
        self.Bind(wx.EVT_CLOSE, self.OnClose)

        self.model = model
        self.thread = None

        psize = parent.GetSize()
        self.SetSize(int(psize.GetWidth() * 0.75), int(psize.GetHeight() * 0.75))

        if prev_result is not None:
            self._res = prev_result
            self.bproblem = prev_result.bproblem
            self.display_bumps()

    def OnRunAnalysis(self, event):
        if self.thread is not None:
            self.run_button.SetLabel("Run Analysis...")
            if bumps.__version__.startswith("0."):
                self.bproblem.fitness.stop_fit = True
            else:
                self.bproblem.stop_fit = True
            return
        self.thread = threading.Thread(target=self.run_bumps)
        self.thread.start()
        self.run_button.SetLabel("Stop Run")

    def get_bumps_param_names(self):
        if bumps.__version__.startswith("0."):
            return list(self.bproblem.model_parameters().keys())
        else:
            return list(self.bproblem.model_parameters()['models'][0].keys())

    def run_bumps(self):
        self.bproblem = self.model.bumps_problem()
        mon = ProgressMonitor(self.bproblem, self.pbar, self.ptxt)
        pop = self.entries["pop"].GetValue()
        burn = self.entries["burn"].GetValue()
        samples = self.entries["samples"].GetValue()
        self.pbar.SetRange(int(samples / (len(self.get_bumps_param_names()) * pop)) + burn)

        with CatchModelError(self, "bumps_modeling") as mgr:
            res = self.model.bumps_fit(
                method="dream",
                pop=pop,
                samples=samples,
                burn=burn,
                thin=1,
                alpha=0,
                outliers="none",
                trim=False,
                monitors=[mon],
                problem=self.bproblem,
            )

        self.run_button.SetLabel("Run Analysis...")
        if mgr.successful:
            self._res = res
            wx.CallAfter(self.display_bumps)

    def display_bumps(self):
        if self.thread is not None:
            self.thread.join(timeout=5.0)
            self.thread = None
        self.pbar.SetValue(0)

        res = self._res
        self.chisq = res.chisq
        if self.chicorrect_checkbox.IsChecked():
            scl = sqrt(self.chisq)
        else:
            scl = 1.0

        self.fom_text.SetLabel("FOM chi²/bars: %.3f" % self.chisq)
        self.draw = res.state.draw()
        pnames = self.get_bumps_param_names()
        sort_indices = [pnames.index(ni) for ni in self.draw.labels]

        self.abs_cov = res.cov
        self.rel_cov = res.cov / res.dx[:, newaxis] / res.dx[newaxis, :]
        rel_max = [0, 1, abs(self.rel_cov[0, 1])]
        if self.normalize_checkbox.IsChecked():
            display_cov = self.rel_cov
            fmt = "%.6f"
        else:
            display_cov = self.abs_cov * scl**2
            fmt = "%.4g"
        self.grid.SetRowLabelValue(0, "Value:")
        self.grid.SetRowLabelValue(1, "Error:")
        for i, ci in enumerate(self.rel_cov):
            plabel = "\n".join(self.draw.labels[i].rsplit("_", 1))
            self.grid.SetColLabelValue(sort_indices[i], plabel)
            self.grid.SetRowLabelValue(sort_indices[i] + 2, plabel)
            self.grid.SetCellValue(0, sort_indices[i], "%.8g" % res.x[i])
            self.grid.SetCellAlignment(0, sort_indices[i], wx.ALIGN_CENTRE, wx.ALIGN_CENTRE)
            self.grid.SetCellValue(1, sort_indices[i], "%.4g" % (res.dx[i] * scl))
            self.grid.SetCellAlignment(1, sort_indices[i], wx.ALIGN_CENTRE, wx.ALIGN_CENTRE)
            self.grid.SetReadOnly(0, sort_indices[i])
            self.grid.SetReadOnly(1, sort_indices[i])
            self.grid.SetCellBackgroundColour(0, sort_indices[i], "#cccccc")
            self.grid.SetCellBackgroundColour(1, sort_indices[i], "#cccccc")
            for j, cj in enumerate(ci):
                self.grid.SetCellValue(sort_indices[i] + 2, sort_indices[j], fmt % display_cov[i, j])
                self.grid.SetReadOnly(sort_indices[i] + 2, sort_indices[j])
                self.grid.SetCellAlignment(sort_indices[i] + 2, sort_indices[j], wx.ALIGN_CENTRE, wx.ALIGN_CENTRE)
                if i == j:
                    self.grid.SetCellBackgroundColour(sort_indices[i] + 2, sort_indices[j], "#888888")
                elif abs(cj) > 0.4:
                    self.grid.SetCellBackgroundColour(sort_indices[i] + 2, sort_indices[j], "#ffcccc")
                elif abs(cj) > 0.3:
                    self.grid.SetCellBackgroundColour(sort_indices[i] + 2, sort_indices[j], "#ffdddd")
                elif abs(cj) > 0.2:
                    self.grid.SetCellBackgroundColour(sort_indices[i] + 2, sort_indices[j], "#ffeeee")
                if i != j and abs(cj) > rel_max[2]:
                    rel_max = [min(i, j), max(i, j), abs(cj)]
        self.hists = _hists(self.draw.points.T, bins=50)
        self.grid.SelectBlock(sort_indices[rel_max[0]]+2, sort_indices[rel_max[1]],
                              sort_indices[rel_max[0]]+2, sort_indices[rel_max[1]], addToSelected=False)

        self.plot_histogram(rel_max[0], rel_max[1])

        # add analysis data to model for later storage in export header
        exdict = {}
        exdict["library"] = "bumps"
        exdict["version"] = bumps_version
        exdict["settings"] = dict(
            pop=self.entries["pop"].GetValue(),
            burn=self.entries["burn"].GetValue(),
            samples=self.entries["samples"].GetValue(),
        )
        exdict["parameters"] = [
            dict(
                name=li,
                value=float(xi),
                error=float(dxi),
                cross_correlations=dict((self.draw.labels[j], float(res.cov[i, j])) for j in range(len(res.x))),
            )
            for i, (li, xi, dxi) in enumerate(zip(self.draw.labels, res.x, res.dx))
        ]
        self.model.extra_analysis["statistics_mcmc"] = exdict
        if (res.dxpm[:, 0] > 0.0).any() or (res.dxpm[:, 1] < 0.0).any():
            ShowInfoDialog(
                self,
                "There is something wrong in the error estimation, low/high values don't have the right sign.\n\n"
                "This can be caused by non single-modal parameter statistics, closeness to bounds or too low value of"
                "'burn' before stampling.\n\n"
                "Using estimated sigma for grid, instead.",
                title="Issue in uncertainty estimation",
            )
            dxpm = array([-res.dx, res.dx]).T * scl
        else:
            dxpm = res.dxpm * scl
        reverse_sort = [sort_indices.index(i) for i in range(len(sort_indices))]
        error_labels = ["(%.3e, %.3e)" % (dxup, dxdown) for dxup, dxdown in dxpm[reverse_sort]]
        self.model.parameters.set_error_pars(error_labels)

    def OnToggleNormalize(self, evt):
        if self.rel_cov is None:
            evt.Skip()
            return
        if self.normalize_checkbox.IsChecked():
            display_cov = self.rel_cov
            fmt = "%.6f"
        else:
            display_cov = self.abs_cov
            if self.chicorrect_checkbox.IsChecked():
                display_cov = display_cov * self.chisq
            fmt = "%.4g"
        pnames = self.get_bumps_param_names()
        sort_indices = [pnames.index(ni) for ni in self.draw.labels]
        for i, ci in enumerate(self.rel_cov):
            for j, cj in enumerate(ci):
                self.grid.SetCellValue(sort_indices[i] + 2, sort_indices[j], fmt % display_cov[i, j])

    def OnToggleChi2(self, evt):
        if self.rel_cov is None:
            evt.Skip()
            return
        if self.chicorrect_checkbox.IsChecked():
            scl = sqrt(self.chisq)
        else:
            scl = 1.0
        pnames = self.get_bumps_param_names()
        sort_indices = [pnames.index(ni) for ni in self.draw.labels]
        res = self._res
        for i, dxi in enumerate(res.dx):
            self.grid.SetCellValue(1, sort_indices[i], "%.4g" % (dxi * scl))

        if (res.dxpm[:, 0] > 0.0).any() or (res.dxpm[:, 1] < 0.0).any():
            dxpm = array([-res.dx, res.dx]).T * scl
        else:
            dxpm = res.dxpm * scl
        reverse_sort = [sort_indices.index(i) for i in range(len(sort_indices))]
        error_labels = ["(%.3e, %.3e)" % (dxup, dxdown) for dxup, dxdown in dxpm[reverse_sort]]
        self.model.parameters.set_error_pars(error_labels)

        if self.normalize_checkbox.IsChecked():
            return
        else:
            display_cov = self.abs_cov
            if self.chicorrect_checkbox.IsChecked():
                display_cov = display_cov * self.chisq
            fmt = "%.4g"
        for i, ci in enumerate(self.rel_cov):
            for j, cj in enumerate(ci):
                self.grid.SetCellValue(sort_indices[i] + 2, sort_indices[j], fmt % display_cov[i, j])

    def OnSelectCell(self, evt):
        ri, rj = evt.GetCol(), evt.GetRow() - 2
        pnames = self.get_bumps_param_names()
        reverse_indices = [self.draw.labels.index(ni) for ni in pnames]
        i = reverse_indices[ri]
        j = reverse_indices[rj]
        if i == j or j < 0:
            return
        elif i > j:
            itmp = i
            i = j
            j = itmp

        self.plot_histogram(i, j)

    def plot_histogram(self, i, j):
        fig = self.plot_panel.figure
        fig.clear()
        ax = fig.add_subplot(111)
        data, x, y = self.hists[(i, j)]
        vmin, vmax = data[data>0].min(), data.max()
        ax.pcolorfast(y, x, maximum(vmin, data), norm=LogNorm(vmin, vmax), cmap="inferno")
        p1 = data.sum(axis=1)
        p1 = 0.5*p1/p1.max()*(y[-1]-y[0])+y[0]
        p1x = (x[:-1]+x[1:])/2.
        p2 = data.sum(axis=0)
        p2 = 0.5*p2/p2.max()*(x[-1]-x[0])+x[0]
        p2y = (y[:-1]+y[1:])/2.
        ax.plot(p1, p1x, color='blue')
        ax.plot(p2y, p2, color='blue')
        ax.set_xlabel(self.draw.labels[j])
        ax.set_ylabel(self.draw.labels[i])
        self.plot_panel.flush_plot()

    def OnClose(self, event):
        if self.thread is not None and self.thread.is_alive():
            # a running bumps simulation can't be interrupted from other thread, stop window from closing
            self.ptxt.SetLabel("Simulation running")
            if bumps.__version__.startswith("0."):
                self.bproblem.fitness.stop_fit = True
            else:
                self.bproblem.stop_fit = True
            self.thread.join(timeout=5.0)
        event.Skip()
